HAREDO(1)

# NAME

haredo - simple and unix-idiomatic build automator

# SYNOPSIS

*haredo* [-vqc] [-j <n>] targets... [++ sources...]

# DESCRIPTION

The process for haredo boils down to this:

```
for each target argument:
	find and run its .do script
	if a .do script cannot be found, give a warning
return true if any argument is newer than their parent, else false.
```

Haredo's .do files are written in plain shell script, and will be run using
_sh -e_ to exit on error.

## Options

[[ -v
:[ print verbose logs - haredo will execute .do files with _sh -ev_ rather
   than just _sh -e_
|  -q
:  (quiet) don't print 'redo' logs
|  -c
:  (colorless) don't print 'redo' logs with ANSI escape codes
|  -f
:  force rebuild - all files in the targets' dependency tree will be rebuilt
|  -j <n>
:  run a maximum of <n> jobs in parallel

## Arguments

haredo interprets each of its arguments as a target, up until the special
argument '++', and then each argument after that is treated as a source file.
If no arguments are provided, haredo interprets this as if it had been called
with the argument "all".

The difference between a target and a source is that haredo will attempt to
build each target using its corresponding .do file, if it can be found, and if
not it will emit a warning. In contrast, sources are assumed not to have a
.do file, and haredo will not try to find it.

Finally, after running the .do script for each target, haredo checks *all*
sources' and targets' "last-modified" timestamp in the filesystem, and returns
a truthy status code if any of them are newer than the parent target. This is
an indicator to the parent .do script that it should rebuild its target.

## Finding a target's .do file

The process for finding a target's .do file is as follows, using the example
target _web/template.html.tmpl_:

```
first check for the existence of web/template.html.tmpl.do.

if not found, starting in the target directory and then moving up to
successive parent directories, check .do files in this order:
	default.html.tmpl.do
	default.tmpl.do
	default.do
```

## Running the .do file

haredo always runs a .do file from the directory it is in. The .do file is
passed three arguments:
. the target path, relative to the .do file being run
. the basename path (target path without any file extensions)
. a temporary path where the target should be written to.

Once the script is complete, the temporary path at _$3_ will be moved to the
target path. This mechanism prevents half-written files sticking around if the
script is interrupted.

The following default values for environment variables are used. They can be
overridden by setting their values before running haredo.

|[ AR
:[ ar
|  AS
:  as
|  CC
:  cc
|  CXX
:  c++
|  HARE
:  hare
|  HAREC
:  harec
|  LD
:  ld
|  QBE
:  qbe
|  SCDOC
:  scdoc


# EXAMPLES

The typical structure of a .do script is as follows:

```
# assemble the dependencies
deps="dep1 dep2"

# run the build on dependencies, and check if they are newer
if haredo $deps
then
	# build the target
fi
```

The above script will only build the target if the result of running `haredo`
on its dependencies gives a truthy result -- which only happens when one of
them is newer.

## A simple C project

Here is a diagram of an example C program, rigged up with haredo .do files.
The program (hello) simply prints the words "Hello, haredo!" to the terminal.
Note: this diagram can be parsed out into a directory using _treecat(1)_
(https://git.sr.ht/~autumnull/treecat).

```
example/
├── all.do:
│ haredo hello || true
│ ./hello
├── build/
│   ├── clean.do:
│   │ rm -f *.o
│   ├── default.o.do:
│   │ src=../src/$2.c
│   │ if haredo ++ $src
│   │ then
│   │     $CC -o $3 -c $src
│   └ fi
├── clean.do:
│ haredo build/clean || true
│ rm -f hello
├── hello.do:
│ deps="build/hello.o build/main.o"
│ if haredo $deps
│ then
│     $CC -o $3 $deps
│ fi
└── src/
    ├── hello.c:
    │ #include <stdio.h>
    │ #include "hello.h"
    │ 
    │ void hello() {
    │     puts("Hello, haredo!");
    │ }
    ├── hello.h:
    │ void hello();
    ├── main.c:
    │ #include "hello.h"
    │ 
    │ void main() {
    │     hello();
    └ }
```

The main entry point for haredo is the *all.do* file. It simply calls "haredo
hello", ignoring whether it's been updated or not ("|| true"), and then runs
the executable.

The primary .do script for "hello" is "hello.do", which exists, so haredo will
execute it. In hello.do we see a more complex use of haredo. We define the
dependencies, then we call haredo on them. This will run the build script for
each .o file, and then return true if either of them is newer than "hello". If
they are, then the "if" block is executed. Inside the if block there is an
environment variable $CC, which defaults to "cc", the system's c compiler.
There is also the argument $3, which is a temporary file to which each .do
file should write its target, and it will then be moved to the target location
after the script is finished, to prevent broken builds. It's technically
possible to write straight to the target path, but this is ill-advised.

Now, each of the .o files referenced in hello.do don't have their own .do
file, so they both default to using "build/default.o.do". In this script, the
$2 argument is used (the basename of the target, i.e. without file extensions)
to find the .c file that the object is built from. Then in the call to haredo,
the special argument '++' is used before the source file name to indicate that
it doesn't have a .do script, since it's a source file. Everything else in
that file should be familiar from hello.do

And that's all! With this setup, the relevant .o file will be rebuilt, along
with the executable, each time one of the source files changes. There is also
a "clean.do", which can be run with "haredo clean" to clean up all the build
files.

# SEE ALSO

sh(1)

# AUTHORS

Maintained by Autumn! <autumnull@posteo.net>, with significant contributions
from Ember Sawady <ecs@d2evs.net>. Up-to-date sources can be found at
https://sr.ht/~autumnull/haredo and bugs/patches can be submitted by email
to ~autumnull/haredo-devel@lists.sr.ht
